@@ -14,25 +14,15 @@
 #include <libxml/tree.h>
 #include <libxml/xpath.h>
 #include <libxml/xpathInternals.h>
-
-/* Libxml2 version < 2.6.15. xmlreader not reliable/present */
-#if LIBXML_VERSION < 20615
-#define USE_XMLREADER 0
-#else
-#define USE_XMLREADER 1
-#endif
-
-#if USE_XMLREADER
 #include <libxml/xmlreader.h>
 #endif
 
-#endif
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
 #include <assert.h>
+#include <limits.h>
 
 #if HAVE_LOCALE_H
 #include <locale.h>
@@ -64,7 +54,7 @@ static void usage(const char *prog)
 {
     fprintf(stderr, "Usage: %s [-i format] [-o format] [-f from] [-t to] "
             "[-l pos=value] [-c cfile] [-s prefix] [-C size] [-n] "
-            "[-p] [-v] [-V] file...\n",
+            "[-p] [-v] [-V] [-O offset] [-L limit] file...\n",
             prog);
 }
 
@@ -107,7 +97,6 @@ static void marcdump_read_line(yaz_marc_t mt, const char *fname)
                 prog, fname, strerror(errno));
         exit(1);
     }
-
     while (yaz_marc_read_line(mt, getbyte_stream,
                               ungetbyte_stream, inf) == 0)
     {
@@ -121,6 +110,11 @@ static void marcdump_read_line(yaz_marc_t mt, const char *fname)
 
 static void marcdump_read_json(yaz_marc_t mt, const char *fname)
 {
+    const char *errmsg;
+    size_t errpos;
+    WRBUF w = wrbuf_alloc();
+    struct json_node *n;
+    int c;
     FILE *inf = fopen(fname, "rb");
     if (!inf)
     {
@@ -128,125 +122,85 @@ static void marcdump_read_json(yaz_marc_t mt, const char *fname)
                 prog, fname, strerror(errno));
         exit(1);
     }
-    else
+    while ((c = getc(inf)) != EOF)
+        wrbuf_putc(w, c);
+    n = json_parse2(wrbuf_cstr(w), &errmsg, &errpos);
+    if (n)
     {
-        const char *errmsg;
-        size_t errpos;
-        WRBUF w = wrbuf_alloc();
-        struct json_node *n;
-        int c;
-
-        while ((c = getc(inf)) != EOF)
-            wrbuf_putc(w, c);
-        n = json_parse2(wrbuf_cstr(w), &errmsg, &errpos);
-        if (n)
+        int r = yaz_marc_read_json_node(mt, n);
+        if (r == 0)
         {
-            int r = yaz_marc_read_json_node(mt, n);
-            if (r == 0)
-            {
-                wrbuf_rewind(w);
-                yaz_marc_write_mode(mt, w);
-                fputs(wrbuf_cstr(w), stdout);
-                wrbuf_rewind(w);
-            }
-            else
-            {
-                fprintf(stderr, "%s: JSON MARC parsing failed ret=%d\n", fname,
-                        r);
-            }
+            wrbuf_rewind(w);
+            yaz_marc_write_mode(mt, w);
+            fputs(wrbuf_cstr(w), stdout);
+            wrbuf_rewind(w);
         }
         else
         {
-            fprintf(stderr, "%s: JSON parse error: %s . pos=%ld\n", fname,
-                    errmsg, (long) errpos);
+            fprintf(stderr, "%s: JSON MARC parsing failed ret=%d\n", fname,
+                    r);
         }
-        wrbuf_destroy(w);
-        fclose(inf);
     }
+    else
+    {
+        fprintf(stderr, "%s: JSON parse error: %s . pos=%ld\n", fname,
+                errmsg, (long) errpos);
+    }
+    wrbuf_destroy(w);
+    fclose(inf);
 }
 
 
 #if YAZ_HAVE_XML2
-static void marcdump_read_xml(yaz_marc_t mt, const char *fname)
+static void marcdump_read_xml(yaz_marc_t mt, const char *fname,
+                              long offset, long limit)
 {
     WRBUF wrbuf = wrbuf_alloc();
-#if USE_XMLREADER
     xmlTextReaderPtr reader = xmlReaderForFile(fname, 0 /* encoding */,
                                                0 /* options */);
-
-    if (reader)
+    int ret;
+    long no = 0;
+    if (reader == 0)
     {
-        int ret;
-        while ((ret = xmlTextReaderRead(reader)) == 1)
-        {
-            int type = xmlTextReaderNodeType(reader);
-            if (type == XML_READER_TYPE_ELEMENT)
-            {
-                char *name = (char *) xmlTextReaderLocalName(reader);
-                if (!strcmp(name, "record") || !strcmp(name, "r"))
-                {
-                    xmlNodePtr ptr = xmlTextReaderExpand(reader);
-
-                    int r = yaz_marc_read_xml(mt, ptr);
-                    if (r)
-                    {
-                        no_errors++;
-                        fprintf(stderr, "yaz_marc_read_xml failed\n");
-                    }
-                    else
-                    {
-                        int write_rc = yaz_marc_write_mode(mt, wrbuf);
-                        if (write_rc)
-                        {
-                            yaz_log(YLOG_WARN, "yaz_marc_write_mode: "
-                                    "write error: %d", write_rc);
-                            no_errors++;
-                        }
-                        fputs(wrbuf_cstr(wrbuf), stdout);
-                        wrbuf_rewind(wrbuf);
-                    }
-                }
-                xmlFree(name);
-            }
-        }
-        xmlFreeTextReader(reader);
+        fprintf(stderr, "%s: cannot open %s:%s\n",
+                prog, fname, strerror(errno));
+        exit(1);
     }
-#else
-    xmlDocPtr doc = xmlParseFile(fname);
-    if (doc)
+    while ((ret = xmlTextReaderRead(reader)) == 1)
     {
-        xmlNodePtr ptr = xmlDocGetRootElement(doc);
-        for (; ptr; ptr = ptr->next)
+        int type = xmlTextReaderNodeType(reader);
+        if (type == XML_READER_TYPE_ELEMENT)
         {
-            if (ptr->type == XML_ELEMENT_NODE)
+            char *name = (char *) xmlTextReaderLocalName(reader);
+            if (!strcmp(name, "record") || !strcmp(name, "r"))
             {
-                if (!strcmp((const char *) ptr->name, "collection"))
+                xmlNodePtr ptr = xmlTextReaderExpand(reader);
+                int r = yaz_marc_read_xml(mt, ptr);
+                if (r)
                 {
-                    ptr = ptr->children;
-                    continue;
+                    no_errors++;
+                    fprintf(stderr, "yaz_marc_read_xml failed\n");
                 }
-                if (!strcmp((const char *) ptr->name, "record") ||
-                    !strcmp((const char *) ptr->name, "r"))
+                else if (no >= offset)
                 {
-                    int r = yaz_marc_read_xml(mt, ptr);
-                    if (r)
+                    int write_rc = yaz_marc_write_mode(mt, wrbuf);
+                    if (write_rc)
                     {
+                        yaz_log(YLOG_WARN, "yaz_marc_write_mode: "
+                                "write error: %d", write_rc);
                         no_errors++;
-                        fprintf(stderr, "yaz_marc_read_xml failed\n");
-                    }
-                    else
-                    {
-                        yaz_marc_write_mode(mt, wrbuf);
-
-                        fputs(wrbuf_cstr(wrbuf), stdout);
-                        wrbuf_rewind(wrbuf);
                     }
+                    fputs(wrbuf_cstr(wrbuf), stdout);
+                    wrbuf_rewind(wrbuf);
                 }
+                no++;
             }
+            xmlFree(name);
         }
-        xmlFreeDoc(doc);
+        if (no - offset >= limit)
+            break;
     }
-#endif
+    xmlFreeTextReader(reader);
     fputs(wrbuf_cstr(wrbuf), stdout);
     wrbuf_destroy(wrbuf);
 }
@@ -256,7 +210,8 @@ static void dump(const char *fname, const char *from, const char *to,
                  int input_format, int output_format,
                  int write_using_libxml2,
                  int print_offset, const char *split_fname, int split_chunk,
-                 int verbose, FILE *cfile, const char *leader_spec)
+                 int verbose, FILE *cfile, const char *leader_spec,
+                 long offset, long limit)
 {
     yaz_marc_t mt = yaz_marc_create();
     yaz_iconv_t cd = 0;
@@ -287,7 +242,7 @@ static void dump(const char *fname, const char *from, const char *to,
     if (input_format == YAZ_MARC_MARCXML || input_format == YAZ_MARC_TURBOMARC || input_format == YAZ_MARC_XCHANGE)
     {
 #if YAZ_HAVE_XML2
-        marcdump_read_xml(mt, fname);
+        marcdump_read_xml(mt, fname, offset, limit);
 #endif
     }
     else if (input_format == YAZ_MARC_LINE)
@@ -302,7 +257,7 @@ static void dump(const char *fname, const char *from, const char *to,
     {
         FILE *inf = fopen(fname, "rb");
         int num = 1;
-        int marc_no = 0;
+        long marc_no;
         int split_file_no = -1;
         if (!inf)
         {
@@ -312,7 +267,7 @@ static void dump(const char *fname, const char *from, const char *to,
         }
         if (cfile)
             fprintf(cfile, "char *marc_records[] = {\n");
-        for(;; marc_no++)
+        for (marc_no = 0L; marc_no - offset < limit; marc_no++)
         {
             const char *result = 0;
             size_t len;
@@ -447,7 +402,7 @@ static void dump(const char *fname, const char *from, const char *to,
 
             if (r == -1)
                 no_errors++;
-            if (r > 0 && result && len_result)
+            if (r > 0 && result && len_result && marc_no >= offset)
             {
                 if (fwrite(result, len_result, 1, stdout) != 1)
                 {
@@ -512,6 +467,8 @@ int main (int argc, char **argv)
     const char *split_fname = 0;
     const char *leader_spec = 0;
     int write_using_libxml2 = 0;
+    long offset = 0L;
+    long limit = LONG_MAX;
 
 #if HAVE_LOCALE_H
     setlocale(LC_CTYPE, "");
@@ -524,7 +481,7 @@ int main (int argc, char **argv)
 
     prog = *argv;
     yaz_enable_panic_backtrace(prog);
-    while ((r = options("i:o:C:npc:xOeXIf:t:s:l:Vv", argv, argc, &arg)) != -2)
+    while ((r = options("i:o:C:npc:xL:O:eXIf:t:s:l:Vv", argv, argc, &arg)) != -2)
     {
         no++;
         switch (r)
@@ -586,10 +543,11 @@ int main (int argc, char **argv)
                     "Use -i marcxml instead\n", prog);
             exit(1);
             break;
+        case 'L':
+            limit = atol(arg);
+            break;
         case 'O':
-            fprintf(stderr, "%s: OAI MARC no longer supported."
-                    " Use MARCXML instead.\n", prog);
-            exit(1);
+            offset = atol(arg);
             break;
         case 'e':
             fprintf(stderr, "%s: -e no longer supported. "
@@ -622,7 +580,7 @@ int main (int argc, char **argv)
             dump(arg, from, to, input_format, output_format,
                  write_using_libxml2,
                  print_offset, split_fname, split_chunk,
-                 verbose, cfile, leader_spec);
+                 verbose, cfile, leader_spec, offset, limit);
             break;
         case 'v':
             verbose++;
